@@ -5,17 +5,18 @@ import torch.nn as nn
 import os
 from torch.autograd import Variable
 import torch.utils.data as Data
+import torch.nn.functional as F
 
 #规范
-#全连接层 Linear：[  "Linear" , [输入数，输出数，是否采用dropout] ]
+#全连接层 Linear：[  "Linear" , [输入数，输出数，bias] ]
 #激活函数：softmax, sigmoid, relu, tanh, elu, prelu
 #nn.BatchNorm2d ------- "batchnorm2d": [  ["batchnorm2d"], [第1维尺寸，分母偏置值，动量，是否需要引入学习参数（使用默认填入None）] ]
-#CNN ： [["CNN"], [输入通道数，输出通道数，核边长，步长，边缘补充长度]]
+#CNN ： [ "CNN" , [输入通道数，输出通道数，核边长，步长，边缘补充长度，bias]]
 #CNN的输入：[batch][通道数][尺寸1][尺寸2]
 
 
-def AddALinearPlease(with_input, and_output, if_dropout=False, nnlist=None):
-    newnn = ["Linear", [with_input, and_output, if_dropout]]
+def AddALinearPlease(with_input, and_output, withbias=True, nnlist=None):
+    newnn = ["Linear", [with_input, and_output, withbias]]
     if istype(nnlist, []):
         nnlist.append(newnn)
         return nnlist
@@ -23,6 +24,29 @@ def AddALinearPlease(with_input, and_output, if_dropout=False, nnlist=None):
         return newnn
 
 
+def AddACNNPlease(input_channels, output_channels, kernal_size, stride, padding_size, bias=True, nnlist=None):
+    newnn = ["CNN", [input_channels, output_channels, kernal_size, stride, padding_size, bias]]
+    if istype(nnlist, []):
+        nnlist.append(newnn)
+        return nnlist
+    else:
+        return newnn
+
+
+def good_dropout(dp):
+    return istype(dp, 0.1) or istype(dp, 1)
+
+def AddADropoutPlease(dropout=None, inplace=False, nnlist=None):
+    dp = 0.0
+    if good_dropout(dropout):
+        dp = dropout
+    if istype(nnlist, []):
+        nnlist.append(["dropout", [dp, inplace]])
+        return nnlist
+    else:
+        return ["dropout", [dp, inplace]]
+
+
 class myNN(nn.Module):
     def __init__(self, nnlist):
         super(myNN, self).__init__()
@@ -34,7 +58,9 @@ class myNN(nn.Module):
             n = nnlist[i]
             if n[0] == "Linear":
                 p = n[1].copy()
-                b = p[2]
+                b = True
+                if len(p) > 2:
+                    b = p[2]
                 layer = nn.Linear(p[0], p[1], b)
                 self.layer.append(layer)
                 self.forward_type.append(0)
@@ -44,19 +70,19 @@ class myNN(nn.Module):
                 hidden_size = p[1]
                 num_layers = p[2]
                 nonlinearity = p[3]
-                if nonlinearity == None:
+                if nonlinearity is None:
                     nonlinearity = "tanh"
                 bias = p[4]
-                if bias == None:
+                if bias is None:
                     bias = True
                 batch_first = p[5]
-                if batch_first == None:
+                if batch_first is None:
                     batch_first = False
-                dropout = p[6]
-                if dropout == None:
-                    dropout = 0
+                dropout = 0
+                if good_dropout(p[6]):
+                    dropout=p[6]
                 bid = p[7]
-                if bid == None:
+                if bid is None:
                     bid = False
                 layer = nn.RNN(
                     input_size=input_size,
@@ -77,11 +103,15 @@ class myNN(nn.Module):
             if n[0] == "CNN": #2
                 #输入：[batch][通道数][尺寸1][尺寸2]
                 p = n[1].copy()
+                b = True
+                if len(p) > 5:
+                    b = p[5]
                 layer = nn.Conv2d(in_channels=p[0],
                                   out_channels=p[1],
                                   kernel_size=p[2],
                                   stride=p[3],
-                                  padding=p[4])
+                                  padding=p[4],
+                                  bias=b)
                 self.layer.append(layer)
                 self.forward_type.append(2)
             if n[0] == "batchnorm2d":
@@ -90,11 +120,11 @@ class myNN(nn.Module):
                 eps = p[1]
                 momentum = p[2]
                 affine = p[3]
-                if eps == None:
+                if eps is None:
                     eps = 1e-05
-                if momentum == None:
+                if momentum is None:
                     momentum = 0.1
-                if affine == None:
+                if affine is None:
                     affine = True
                 layer = nn.BatchNorm2d(num, eps, momentum, affine)
                 self.layer.append(layer)
@@ -112,34 +142,55 @@ class myNN(nn.Module):
                     d = n[1]
                 self.layer.append(nn.Softmax(dim=d))
                 self.forward_type.append(-1)
+            if n[0] == "tanh":
+                self.layer.append(nn.Tanh())
+                self.forward_type.append(-1)
             if n[0] == "elu":
                 self.layer.append(nn.ELU())
                 self.forward_type.append(-1)
             if n[0] == "prelu":
-                if len(n) == 1 or n[1][0] == None:
+                if len(n) == 1 or n[1][0] is None:
                     nump = 1
                 else:
                     nump = n[1][0]
-                if len(n) == 1 or n[1][1] == None:
+                if len(n) == 1 or n[1][1] is None:
                     initialize = 0.25
                 else:
                     initialize = n[1][1]
                 self.layer.append(nn.PReLU(num_parameters=nump, init=initialize))
                 self.forward_type.append(-1)
+            if n[0] == "maxpool2d":
+                self.forward_type.append(3)
+                self.layer.append(nn.MaxPool2d(kernel_size=n[1]))
+            if n[0] == "dropout":
+                p = n[1]
+                self.forward_type.append(50)
+                if len(p>1):
+                    self.layer.append(nn.Dropout(p[0], p[1]))
+                else:
+                    self.layer.append(nn.Dropout(p[0]))
         self.layer_module = nn.ModuleList(self.layer)
 
-    def forward(self, x, H=None):
+    def forward(self, x, H=None, start_layer=0, stop_layer=None):
         i = 0
         numh = 0
         for layer in self.layer_module:
+            if i < start_layer:
+                i += 1
+                continue
+            #直接跳过前几层，这在没有设置start_layer时是安全的，因为一开始就有i==0==start_layer
+            #下列是正式的前向传播
             if self.forward_type[i] == 0 \
                 or self.forward_type[i] == -1\
-                    or self.forward_type[i] == 2:
+                or self.forward_type[i] == 2\
+                or self.forward_type[i] == 3\
+                or self.forward_type[i] == 50:
                 x = layer(x)
+            #RNN有隐层输入，与CNN的输入格式不同
             if self.forward_type[i] == 10 or self.forward_type[i] == 11:
                 if self.forward_type[i] == 10:
                     #batch first, [batch][time][dim]
-                    if H == None or len(H) <= numh or H[numh] == None:
+                    if H is None or len(H) <= numh or H[numh] is None:
                         out = layer(x)[0]
                     else:
                         out = layer(x, H[numh])[0]
@@ -148,7 +199,7 @@ class myNN(nn.Module):
                 if self.forward_type[i] == 11:
                     #不是batch first, [time][batch][dim]
                     x = x.transpose(0, 1)
-                    if H == None or len(H) <= numh or H[numh] == None:
+                    if H is None or len(H) <= numh or H[numh] is None:
                         out = layer(x)[0]
                     else:
                         out = layer(x, H[numh])[0]
@@ -156,6 +207,8 @@ class myNN(nn.Module):
                 numh += 1
 
             i += 1
+            if i == stop_layer:
+                break
         return x
 
 #需要自己写的新函数：
@@ -163,14 +216,18 @@ class myNN(nn.Module):
 #训练的过程，如何存取数据，如何测试
 #对接到具体场景的方法
 class myNetwork:
-    def __init__(self, nnlist, bs=1, lr=0.01):
+    def __init__(self, nnlist, batch_size=1, lr=0.01):
         self.model = myNN(nnlist)
         self.criterion = nn.MSELoss(reduction='mean')
         self.optimizer = torch.optim.Adam(self.model.parameters(), lr)
         self.training_loss_list = []
         self.testing_loss_list = []
         self.testing_accuracy_list = []
-        self.batch_size = bs
+        self.batch_size = batch_size
+
+    def to(self, device):
+        self.model = self.model.to(device)
+        self.criterion = self.criterion.to(device)
 
     def set_optimizer(self, t, lr):
         if t == "Adam":
@@ -226,11 +283,14 @@ class myNetwork:
             return Variable(torch.tensor(x, dtype=dtype))
 
     def training_of_single_batch(self, x, y):
+        self.model.training = True
         pred_y = self.forward(x)
         loss = self.loss(pred_y, y)
         self.optimizer.zero_grad()
         loss.backward()
         self.optimizer.step()
+        self.model.training = False
+        del pred_y
         return loss
 
 
@@ -257,6 +317,22 @@ def AddAReluPlease(nnlist=None):
     else:
         return ["relu"]
 
+def AddATanhPlease(nnlist=None):
+    if istype(nnlist, []):
+        nnlist.append(["tanh"])
+        return nnlist
+    else:
+        return ["tanh"]
+
+
+def AddAMaxPoolPlease(kernel_size=2, nnlist=None):
+    s = ["maxpool2d", kernel_size]
+    if istype(nnlist, []):
+        nnlist.append(s)
+        return nnlist
+    else:
+        return s
+
 
 
 
